#include <stdint.h>

/**
 * Enumeration for different possible errors
 */
typedef enum tsc_result {
  /**
   * Function did not have any failures
   */
  TSC_RESULT_OK = 1000,
  /**
   * A URL with an invalid format was used
   */
  TSC_RESULT_INVALID_URL = 1001,
  /**
   * An input string where UTF8 is required failed validation
   */
  TSC_RESULT_UTF8_VALIDATION_FAILURE = 1002,
  /**
   * The HTTP connection exceeded the max redirects value
   */
  TSC_RESULT_TOO_MANY_REDIRECTS = 1003,
  /**
   * The HTTP connection exceeded the connect or connection timeout
   */
  TSC_RESULT_TIMED_OUT = 1004,
  /**
   * A non-200 HTTP response was returned
   */
  TSC_RESULT_RESPONSE_ERROR = 1005,
  /**
   * An unexpected error was found
   */
  TSC_RESULT_UNEXPECTED_ERROR = 1006,
  /**
   * The DNS lookup failed
   */
  TSC_RESULT_DNS_LOOKUP_FAILURE = 1007,
  /**
   * The connection failed to be established
   */
  TSC_RESULT_CONNECTION_FAILED = 1008,
  /**
   * The connection was aborted
   */
  TSC_RESULT_CONNECTION_ABORTED = 1009,
  /**
   * The server refused the connection
   */
  TSC_RESULT_CONNECTION_REFUSED = 1010,
  /**
   * The server reset the connection
   */
  TSC_RESULT_CONNECTION_RESET = 1011,
  /**
   * A bad header was returned from the server
   */
  TSC_RESULT_BAD_HEADER = 1012,
  /**
   * A bad status was returned from the server
   */
  TSC_RESULT_BAD_STATUS = 1013,
  /**
   * An attempt was made to use a proxy but it is not supported
   */
  TSC_RESULT_NO_PROXY_SUPPORT = 1014,
  /**
   * An attempt was made using http, but client only accepts https
   */
  TSC_RESULT_NO_HTTPS = 1015,
} tsc_result;

/**
 * Struct for HTTP requests
 */
typedef struct tsc_http_request tsc_http_request;

/**
 * Struct for the HTTP response
 */
typedef struct tsc_http_response tsc_http_response;

/**
 * Struct for HTTP sessions
 */
typedef struct tsc_http_session tsc_http_session;

const char *tsc_strerror(enum tsc_result error);

/**
 * Issue the request to the network and get back a response. This function call
 * will block until the request has succeeded, failed, or timed out. The status
 * of the call can be read from the return value. If the function call returns
 * TSC_RESULT_OK then the response is valid, otherwise the response should be
 * not be used.
 *
 * @param request A pointer to a request which was returned from
 *                `tsc_http_session_get()`, `tsc_http_session_put()`,
 *                `tsc_http_session_patch()`, `tsc_http_session_delete()`,
 *                or `tsc_http_session_post()`.
 * @param response An unallocated pointer to a response struct. This function
 *                 will allocate the response struct. After use the response
 *                 struct should be unallocated by calling
 *                 `tsc_http_response_free()`. The struct will only be
 *                 allocated if the function is successful.
 * @return If the request fails for some reason it will be return here.
 *         Otherwise a value of TSC_RESULT_OK will be returned.
 */
enum tsc_result tsc_http_request_send(struct tsc_http_request *request,
                                      struct tsc_http_response **response);

/**
 * Issue the request to the network and get back a response. This request will
 * upload data with the request. The 'Content-Length' header is implicitly set
 * to data_length. This function call will block until the request has succeeded,
 * failed, or timed out. The status of the call can be read from the return
 * value. If the function call returns TSC_RESULT_OK then the response is valid,
 * otherwise the response should be not be used.
 *
 * @param request A pointer to a request which was returned from
 *                `tsc_http_session_get()`, `tsc_http_session_put()`,
 *                `tsc_http_session_patch()`, `tsc_http_session_delete()`,
 *                or `tsc_http_session_post()`.
 * @param data A pointer to the data to send. This is often a file or JSON document.
 * @param data_length The length of the buffer pointed to by data.
 * @param response An unallocated pointer to a response struct. This function
 *                 will allocate the response struct. After use the response
 *                 struct should be unallocated by calling
 *                 `tsc_http_response_free()`. The struct will only be
 *                 allocated if the function is successful.
 * @return If the request fails for some reason it will be return here.
 *         Otherwise a value of TSC_RESULT_OK will be returned.
 */
enum tsc_result tsc_http_request_send_bytes(struct tsc_http_request *request,
                                            const uint8_t *data,
                                            uintptr_t data_length,
                                            struct tsc_http_response **response);

/**
 * Issue the request to the network and get back a response. This request will
 * upload data with the request. The 'Content-Length' header is implicitly set
 * to data_length. This function call will block until the request has succeeded,
 * failed, or timed out. The status of the call can be read from the return
 * value. If the function call returns TSC_RESULT_OK then the response is valid,
 * otherwise the response should be not be used.
 *
 * @param request A pointer to a request which was returned from
 *                `tsc_http_session_get()`, `tsc_http_session_put()`,
 *                `tsc_http_session_patch()`, `tsc_http_session_delete()`,
 *                or `tsc_http_session_post()`.
 * @param name A pointer to a null terminated string which containing the name of the HTTP header to set
 * @param value A pointer to a null terminated string containing the value of the HTTP header to set
 * @return If the name or value is not UTF-8 encoded an error will be returned. Otherwise a value of TSC_RESULT_OK will be returned on success.
 */
enum tsc_result tsc_http_request_set_header(struct tsc_http_request *request,
                                            const char *name,
                                            const char *value);

/**
 * Get the current headers set for the request
 * @param request A pointer to a request which was returned from
 *                `tsc_http_session_get()`, `tsc_http_session_put()`,
 *                `tsc_http_session_patch()`, `tsc_http_session_delete()`,
 *                or `tsc_http_session_post()`.
 * @param callback A function pointer which will be called for each key/value
 *                 pair along with the corresponding sizes and a pointer to
 *                 'userdata'
 * @param userdata A pointer to allow access to shared data operated on by the
 *                 callback function.
 * @return A pointer to the internal buffer holding the response. The buffer
 *         will be deallocated once `tsc_http_response_free()` is called.
 */
enum tsc_result tsc_http_request_get_headers(const struct tsc_http_request *request,
                                             void (*callback)(const char*, size_t, const char*, size_t, void*),
                                             void *userdata);

/**
 * Deallocate a request. This function should be called when the request
 * is no longer in use.
 *
 * @param request A pointer to a request which was returned from
 *                `tsc_http_session_get()`, `tsc_http_session_put()`,
 *                `tsc_http_session_patch()`, `tsc_http_session_delete()`,
 *                or `tsc_http_session_post()`.
 */
void tsc_http_request_free(struct tsc_http_request *request);

/**
 * Get the HTTP status code from the response.
 * @param response A pointer to an http_response returned from
 *                 `tsc_http_request_send()`
 * @return An integer value for the HTTP status code returned.
 *
 */
uint16_t tsc_http_response_get_status(const struct tsc_http_response *response);

/**
 * Get the status text from the response. This function will return a pointer
 * to the internal buffer holding the status text.
 * /// @param response A pointer to an http_response returned from
 *                 `tsc_http_request_send()`
 * @return A pointer to the internal buffer holding the status text. This will become invalid once the response is freed.
 */
const unsigned char *tsc_http_response_get_status_text(const struct tsc_http_response *response);

/**
 * Get the body of the response. This function will return a pointer to the
 * internal buffer holding the response.
 * @param response A pointer to an http_response returned from
 *                 `tsc_http_request_send()`
 * @param length Returns the length in bytes of the body
 * @return A pointer to the internal buffer holding the response. The buffer
 *         will be deallocated once `tsc_http_response_free()` is called.
 * @note The returned pointer is *not* null terminated (i.e. it is not a string).
 *
 */
const unsigned char *tsc_http_response_body(const struct tsc_http_response *response,
                                            uintptr_t *len);

/**
 * Deallocate the response once it is no longer in use. Any pointers to data
 * within the response will become invalid.
 * @param response A pointer to an http_response returned from
 *                 `tsc_http_request_send()` or `tsc_http_request_send_bytes()`
 *
 */
void tsc_http_response_free(struct tsc_http_response *response);

/**
 * Get the headers in the response. The callback function will be called for
 * each header and will be passed a pointer to userdata.
 *
 * @param response A pointer to an http_response returned from
 *                 `tsc_http_request_send()`
 * @param callback A function pointer which will be called for each key/value
 *                 pair along with the corresponding sizes and a pointer to
 *                 'userdata'
 * @param userdata A pointer to allow access to shared data operated on by the
 *                 callback function.
 * @return A pointer to the internal buffer holding the response. The buffer
 *         will be deallocated once `tsc_http_response_free()` is called.
 */
enum tsc_result tsc_http_response_get_headers(const struct tsc_http_response *response,
                                              void (*callback)(const char*, size_t, const char*, size_t, void*),
                                              void *userdata);

/**
 * Create a new secure HTTP session. This function will return a pointer to an
 * internally allocated struct for managing the HTTP session. This struct will
 * manage settings and internal memory usage. The first call to this function
 * may initialize some globally shared internal memory.
 *
 * @return A pointer to an internally allocated struct. This pointer must be
 *         freed by being passed to `tsc_http_session_free` when no longer in
 *         use.
 */
struct tsc_http_session *tsc_http_session_new(void);

/**
 * Set timeout for all connections in this session. Note: this does not include the connect timeout.
 * @param session A pointer to session which was allocated from
 *                `tsc_http_session_new()`.
 * @param timeout The length of time in seconds allowed before a connection will timeout.
 *
 */
void tsc_http_session_set_timeout(struct tsc_http_session *session, uint64_t timeout);

/**
 * Set the connect timeout for all connections in this session. Note: this only includes the initial
 * connect timeout, not the timeout for the overall connection.
 * @param session A pointer to session which was allocated from
 *                `tsc_http_session_new()`.
 * @param timeout The length of time in seconds allowed before a connection will timeout.
 *
 */
void tsc_http_session_set_timeout_connect(struct tsc_http_session *session, uint64_t timeout);

/**
 * Set the max number of HTTP redirects for this session.
 * @param session A pointer to session which was allocated from
 *                `tsc_http_session_new()`.
 * @param redirects The max number of redirects allowed
 *
 */
void tsc_http_session_set_max_redirects(struct tsc_http_session *session, uint32_t redirects);

/**
 * The user agent string will be copied so it can be freed after use
 * @param session A pointer to session which was allocated from
 *                `tsc_http_session_new()`.
 * @param user_agent A string to specify the user agent.
 */
void tsc_http_session_set_user_agent(struct tsc_http_session *session, const char *user_agent);

/**
 * Free a secure HTTP session. This function will deallocate the tsc_http_session
 * and any allocated requests or responses that have not already been freed.
 *
 * @param session A pointer to session which was allocated from
 *                `tsc_http_session_new()`.
 */
void tsc_http_session_free(struct tsc_http_session *session);

/**
 * Set the URL future requests will use. The URL must start with HTTPS. All
 * requests made after calling this function will use this URL until this
 * function is called again.
 *
 * @param session A pointer to a session which was allocated from
 *                `tsc_http_session_new()`
 * @param url A pointer to a valid URL starting with https://.
 * @return If the URL fails to be parsed this function will return an error.
 *         Otherwise TSC_RESULT_OK will be returned.
 */
enum tsc_result tsc_http_session_set_url(struct tsc_http_session *session, const char *url);

/**
 * Create an HTTP GET request. This type of request is often used to request
 * data from a server.
 * @param session A pointer to a session which was allocated from
 *                `tsc_http_session_new()`
 * @return A pointer to an internally allocated `tsc_http_request` structure.
 *         This request can be freed by calling `tsc_http_request_free()`.
 */
struct tsc_http_request *tsc_http_session_get(struct tsc_http_session *session);

/**
 * Create an HTTP PUT request. This type of request is often used to send
 * data to a server in an idempotent way.
 * @param session A pointer to a session which was allocated from
 *                `tsc_http_session_new()`
 * @return A pointer to an internally allocated `tsc_http_request` structure.
 *         This request can be freed by calling `tsc_http_request_free()`.
 */
struct tsc_http_request *tsc_http_session_put(struct tsc_http_session *session);

/**
 * Create an HTTP POST request. This type of request is often used to send
 * data to a server.
 * @param session A pointer to a session which was allocated from
 *                `tsc_http_session_new()`
 * @return A pointer to an internally allocated `tsc_http_request` structure.
 *         This request can be freed by calling `tsc_http_request_free()`.
 */
struct tsc_http_request *tsc_http_session_post(struct tsc_http_session *session);

/**
 * Create an HTTP DELETE request. This type of request is often used to delete
 * a resource on a server.
 * @param session A pointer to a session which was allocated from
 *                `tsc_http_session_new()`
 * @return A pointer to an internally allocated `tsc_http_request` structure.
 *         This request can be freed by calling `tsc_http_request_free()`.
 */
struct tsc_http_request *tsc_http_session_delete(struct tsc_http_session *session);

/**
 * Create an HTTP PATCH request. This type of request is often used to modify
 * a resource on a server.
 * @param session A pointer to a session which was allocated from
 *                `tsc_http_session_new()`
 * @return A pointer to an internally allocated `tsc_http_request` structure.
 *         This request can be freed by calling `tsc_http_request_free()`.
 */
struct tsc_http_request *tsc_http_session_patch(struct tsc_http_session *session);
